/*=============================================================================
gdmalloc - Memory leak detection helper utility
gdmalloc.c - CTOR hook
  Copyright (C) Jan "Hirogen2" Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  http://linux01.org:2222/f/hxtools.tbz2
=============================================================================*/

/*-----------------------------------------------------------------------------
GDmalloc is intended to be as free as possible. You do not need to do anything
else than passing -lgdmalloc -rdynamic (and probably -g ;-) to gcc for your
executable. No CFLAGS, no define, no include, no alternate names. Nothing.
Even more: you can set the environment variable LD_PRELOAD to ./libgdmalloc.so
to memory-debug about everything, even binary-only software (which does NOT
call internal, statically-linked malloc, calloc, realloc and free, if
applicable.)

And thus, GDmalloc only compiles with GCC.

Options are set via the environment variable GDMALLOC. Current flags:
  C     do not use ANSI color
  D     deactive gdmalloc
-----------------------------------------------------------------------------*/
#define _GNU_SOURCE 1
#include <sys/types.h>
#include <bfd.h>
#include <ctype.h>
#include <dlfcn.h>
#include <malloc.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX.h>

typedef unsigned char byte;
typedef unsigned long Ulong;

/*=============================================================================
  GDmalloc has a copy of libHX's btree.c, renamed to not clash with the
  original.

libHX:
  Copyright (C) Jan "Hirogen2" Engelhardt <jengelh [at] linux01 gwdg de>,
  1999 - 2003
  -- libHX distributed under the Frontier Artistic License or the GNU General
  -- Public License. See doc/FAL1.txt and doc/GPL2.txt for details.
-----------------------------------------------------------------------------*/
  enum {
    PATH_LEFT = 0,
    PATH_RIGHT,
    NODE_UNUSED = 0,
    NODE_BLACK,
    NODE_RED,
    BT_MAXDEP = 36,
  };

#ifdef __cplusplus
  extern "C" {
#endif

  enum {
    GDMBT_CPTR = 1 << 1, // copy pointer contents (instead of just linking them)
    GDMBT_LKEY = 1 << 2, // link keys (instead of copying them)
    GDMBT_HASH = 1 << 3, // activates associative array behavior
    GDMBT_PCMP = 1 << 4, // pointer to comparison routine is passed
  };

  struct GDMbtree_node {
    struct GDMbtree_node *st[2];
    void *key, *data;
    unsigned char color;
  };

  struct GDMbtree {
    struct GDMbtree_node *tree;
    unsigned long opts;
    unsigned long itemcount, ta;
    int (*pcmp)(void *, void *);
    void *uptr;
  };

  struct GDMbtrav {
    struct GDMbtree *btree;
    struct GDMbtree_node *cnode;
    void *cid;
    unsigned long tree_id;

    struct GDMbtree_node *path[BT_MAXDEP];
    byte dir[BT_MAXDEP], depth;
  };

  extern struct GDMbtree *GDMbtree_init(unsigned long, ...);
  extern struct GDMbtree_node *GDMbtree_add(struct GDMbtree *, void *, ...);
  extern struct GDMbtree_node *GDMbtree_find(struct GDMbtree *, void *);
  extern void *GDMbtree_get(struct GDMbtree *, void *);
  extern void *GDMbtree_del(struct GDMbtree *, void *);
  extern struct GDMbtrav *GDMbtrav_init(struct GDMbtree *, void *);
  extern struct GDMbtree_node *GDMbtraverse(struct GDMbtrav *);
  extern void GDMbtrav_free(struct GDMbtrav *);
  extern void GDMbtree_free(struct GDMbtree *);

#ifdef __cplusplus
  }
#endif
//=============================================================================

#define MAXFNLEN 1024
#define MAXLNLEN 1024
#define SBT_SIZE 256 /* maximum buffer size for stack backtrace dumps */
#define _weak_alias(name, aname) \
  extern __typeof(name) aname __attribute__((weak, alias(#name)));
#define _strong_alias(name, aname) \
  extern __typeof(name) aname __attribute__((alias(#name)));

// Function prototypes
extern void *__libc_malloc(size_t);
extern void *__libc_calloc(size_t, size_t);
extern void *__libc_realloc(void *, size_t);
extern void __libc_free(void *);

extern void _GDM_init(void *) __attribute__((constructor));
extern void _GDM_deinit(void) __attribute__((destructor));
extern void *_GDM_malloc(size_t);
extern void *_GDM_calloc(size_t, size_t);
extern void *_GDM_realloc(void *, size_t);
extern void _GDM_free(void *);
extern void *_GDM_malloc_hook(size_t, const void *);
extern void *_GDM_realloc_hook(void *, size_t, const void *);
extern void _GDM_free_hook(void *, const void *);

static void _GDM_dump_unfreed(void);
static int _GDM_flag(char);
static void _GDM_hooks_deploy(void);
static void _GDM_hooks_restore(void);
static int _GDM_load_file(const char *, unsigned long);
static int _GDM_print(char, char *, ...);
static const char *_GDM_rsymbol(const void *);
static char *_GDM_stack_backtrace(void);
static void _GDM_update_maps(void);
static int _GDM_value_cmp(void *, void *);

// Variables
static void *(*ORIG_hmalloc)(size_t, const void *) = NULL;
static void *(*ORIG_hrealloc)(void *, size_t, const void *) = NULL;
static void (*ORIG_hfree)(void *, const void *) = NULL;
static pthread_mutex_t ext_lock = PTHREAD_MUTEX_INITIALIZER;
static struct GDMbtree *_GDM_aobj = NULL, *_GDM_map = NULL, *_GDM_stab = NULL;
static char *flags = NULL;
static int _GDM_inited = 0;

//-----------------------------------------------------------------------------
void _GDM_init(void *caller) {
    if(_GDM_inited) { return; }
    pthread_mutex_init(&ext_lock, NULL);
    pthread_mutex_lock(&ext_lock);
    if(_GDM_inited++) { return; }

    flags = getenv("GDMALLOC");
    if(!_GDM_flag('A')) { return; }

    _GDM_aobj = GDMbtree_init(GDMBT_HASH | GDMBT_LKEY | GDMBT_PCMP, _GDM_value_cmp);
    _GDM_map = GDMbtree_init(GDMBT_CPTR | GDMBT_PCMP, strcmp);
    _GDM_stab = GDMbtree_init(GDMBT_HASH | GDMBT_LKEY | GDMBT_CPTR | GDMBT_PCMP, _GDM_value_cmp);

    ORIG_hmalloc = __malloc_hook;
    ORIG_hrealloc = __realloc_hook;
    ORIG_hfree = __free_hook;

    _GDM_print(0, "GDmalloc ready\n");
    pthread_mutex_unlock(&ext_lock);
    _GDM_hooks_deploy();
    return;
}

void _GDM_deinit(void) {
    if(!_GDM_flag('A')) { return; }
    _GDM_hooks_restore();
    _GDM_print(0, "GDmalloc hooks deinitialized\n");
    if(_GDM_aobj->itemcount > 0) { _GDM_dump_unfreed(); }
    GDMbtree_free(_GDM_aobj);
    GDMbtree_free(_GDM_map);
    GDMbtree_free(_GDM_stab);
    pthread_mutex_destroy(&ext_lock);
    return;
}

//-----------------------------------------------------------------------------
// Re-wrappers
void *_GDM_malloc(size_t s) {
    return __libc_malloc(s);
}

void *_GDM_calloc(size_t n, size_t s) {
    return __libc_calloc(n, s);
}

void *_GDM_realloc(void *ptr, size_t s) {
    return __libc_realloc(ptr, s);
}

void _GDM_free(void *ptr) {
    __libc_free(ptr);
    return;
}

//-----------------------------------------------------------------------------
// Libc hooks
void *_GDM_malloc_hook(size_t s, const void *caller) {
    void *rv;

    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();
    _GDM_update_maps();

    if((rv = __libc_malloc(s)) != NULL) {
        GDMbtree_add(_GDM_aobj, rv, _GDM_stack_backtrace());
    }

    _GDM_print('v', "[objs=%5ld] 0x%08lX = malloc(s=0x%08lX)\n",
     _GDM_aobj->itemcount, (Ulong)rv, (Ulong)s);

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return rv;
}

void *_GDM_realloc_hook(void *ptr, size_t s, const void *caller) {
    void *rv;

    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();

    rv = __libc_realloc(ptr, s);
    if(ptr != rv) {
        if(ptr != NULL) { GDMbtree_del(_GDM_aobj, ptr); }
        if(rv != NULL) { GDMbtree_add(_GDM_aobj, rv, _GDM_stack_backtrace()); }
    }

    _GDM_print('v', "[objs=%5ld] 0x%08lX = realloc(ptr=0x%08lX, s=0x%08lX)\n",
     _GDM_aobj->itemcount, (Ulong)rv, (Ulong)ptr, (Ulong)s);

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return rv;
}

void _GDM_free_hook(void *ptr, const void *caller) {
    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();
    __libc_free(ptr);

    GDMbtree_del(_GDM_aobj, ptr);
    _GDM_print('v', "[objs=%5ld] free(ptr=0x%08lX)\n",
     _GDM_aobj->itemcount, (Ulong)ptr);

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return;
}

//-----------------------------------------------------------------------------
// Internal functions
static void _GDM_dump_unfreed(void) {
    struct GDMbtrav *travp = GDMbtrav_init(_GDM_aobj, NULL);
    struct GDMbtree_node *n;

    while((n = GDMbtraverse(travp)) != NULL) {
      _GDM_print(0, "Unfreed object 0x%08lX (BT: %s)\n",
       (Ulong)n->key, n->data);
    }

    GDMbtrav_free(travp);
    return;
}

static int _GDM_flag(char f) {
    if(f == 0 || (flags != NULL && strchr(flags, f) != NULL)) { return 1; }
    return 0;
}

static void _GDM_hooks_deploy(void) {
    __malloc_hook = _GDM_malloc_hook;
    __realloc_hook = _GDM_realloc_hook;
    __free_hook = _GDM_free_hook;
    return;
}

static void _GDM_hooks_restore(void) {
    __malloc_hook = ORIG_hmalloc;
    __realloc_hook = ORIG_hrealloc;
    __free_hook = ORIG_hfree;
    return;
}

static int _GDM_load_file(const char *filename, unsigned long base) {
    FILE *fp;
    char buf[MAXLNLEN];

    _GDM_print(0, "Loading symbol table from %s\n", filename, base);
    snprintf(buf, MAXLNLEN, "nm %s 2>/dev/null", filename);
    if((fp = popen(buf, "r")) == NULL) { return 0; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *addrp = buf, *type, *name;
        unsigned long addr;

        if(*addrp == ' ') { continue; }
        if((type = strchr(buf, ' ')) == NULL) { continue; }
        *type++ = '\0';
        if(*type != 't' && *type != 'T') { continue; }
        if((name = strchr(type, ' ')) == NULL) { continue; }
        *name++ = '\0';

        HX_chomp(name);
        addr = strtoul(addrp, NULL, 16);
        if(addr < base) { addr += base; }
        _GDM_print(0, "Adding symbol %s (%p)\n", name, addr);
        GDMbtree_add(_GDM_stab, (void *)addr, name);
        malloc(32);
    }

    pclose(fp);
    return 1;
}

static int _GDM_print(char f, char *fmt, ...) {
    int rv = 0;
    va_list argp;
    va_start(argp, fmt);
    if(_GDM_flag(f)) {
        if(!_GDM_flag('C')) { fprintf(stderr, "\e[35mGDM\e[1m>\e[0m "); } // ]]]
        else { fprintf(stderr, "GDM> "); }
        rv = vfprintf(stderr, fmt, argp);
    }
    va_end(argp);
    return rv;
}

static const char *_GDM_rsymbol(const void *addr) {
    Dl_info q;
    if(!dladdr(addr, &q)) { return NULL; }
    return q.dli_sname;
/*
    int v;
    char *sym = NULL;
    void *bp = NULL;
    struct GDMbtree *btree = _GDM_stab;
    struct GDMbtree_node *node = btree->tree;

    while(node != NULL) {
        v = btree->pcmp((void *)addr, node->key);
        if(v == 0) { sym = node->data; bp=node->key;break; }
        v = v > 0;
        if(v == 0) { sym = node->data; bp = node->key; }
        node = node->st[v];
    }

    print(0, "Best symbol for %p: %p=%s\n", addr, bp, sym);
    return sym;
*/
    return NULL;
}

static char *_GDM_stack_backtrace(void) {
    void *ebp;
    char buf[SBT_SIZE], ibuf[256];

    asm volatile("movl %%ebp, %0;" :"=m"(ebp));
    memset(buf, '\0', SBT_SIZE);

//    fprintf(stderr, "Trace:\n");
    while(ebp != 0x0) {
      void *f_addr = *(void **)(ebp + sizeof(void *));
      const char *sym = _GDM_rsymbol(f_addr);
/*
      _GDM_print(0, "EBP=0x%08lX, f_addr=0x%08lX, sym=%s\n",
       (unsigned long)ebp, (unsigned long)f_addr, sym);
*/
      if(sym == NULL) {
          snprintf(ibuf, 256, "0x%08lX", (unsigned long)f_addr);
      } else {
          snprintf(ibuf, 256, "%s", sym);
      }

      HX_strlcat(buf, ibuf, SBT_SIZE);
      ebp = *(void **)ebp;

      if(ebp != 0x0) { HX_strlcat(buf, " < ", SBT_SIZE); }
    }

    return HX_strdup(buf);
}

static void _GDM_update_maps(void) {
    FILE *fp;
    char buf[MAXLNLEN], fbuf[MAXFNLEN];
    snprintf(fbuf, MAXFNLEN, "/proc/%d/maps", getpid());
    if((fp = fopen(fbuf, "r")) == NULL) { return; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *base = buf;
        char *filename = strchr(buf, '/');
        if(filename == NULL) { continue; }
        HX_chomp(buf);
        *strchr(buf, '-') = '\0';
        if(GDMbtree_find(_GDM_map, filename) == NULL) {
            _GDM_load_file(filename, strtoul(base, NULL, 16));
            GDMbtree_add(_GDM_map, filename);
        }
    }

    fclose(fp);
    return;
}

static int _GDM_value_cmp(void *pa, void *pb) {
    return (pa > pb) - (pa < pb);
}

//-----------------------------------------------------------------------------
// Redirect any malloc() call to us, then __libc_malloc()
_strong_alias(_GDM_malloc,  malloc);
_strong_alias(_GDM_calloc,  calloc);
_strong_alias(_GDM_realloc, realloc);
_strong_alias(_GDM_free,    free);

/*=============================================================================
libHX/btree.c, modified
  Copyright (C) Jan "Hirogen2" Engelhardt <jengelh [at] linux01 gwdg de>,
  1999 - 2003
  -- libHX distributed under the Frontier Artistic License or the GNU General
  -- Public License. See doc/FAL1.txt and doc/GPL2.txt for details.
=============================================================================*/
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

static void GDMbtrav_cid(struct GDMbtrav *, struct GDMbtree_node *);
static struct GDMbtree_node *GDMbtrav_next(struct GDMbtrav *);
static struct GDMbtree_node *GDMbtrav_rewalk(struct GDMbtrav *);
static void GDMbtree_free_dive(struct GDMbtree *, struct GDMbtree_node *);
static int GDMbtree_value_cmp(void *, void *);

//-----------------------------------------------------------------------------
struct GDMbtree *GDMbtree_init(unsigned long opts, ...) {
  struct GDMbtree *btree;
  va_list argp;
  va_start(argp, opts);

  if((btree = malloc(sizeof(struct GDMbtree))) == NULL) { return NULL; }
  memset(btree, 0, sizeof(struct GDMbtree));
  btree->opts = opts;
  btree->itemcount = 0;
  btree->ta = 1;

  if(opts & GDMBT_PCMP) { btree->pcmp = va_arg(argp, void *); }
  else if(opts & GDMBT_HASH) { btree->pcmp = (void *)strcmp; }
  else { btree->pcmp = GDMbtree_value_cmp; }

  va_end(argp);
  return btree;
}

struct GDMbtree_node *GDMbtree_add(struct GDMbtree *btree, void *key, ...) {
  struct GDMbtree_node **tree, *node;
  struct GDMbtree_node *path[BT_MAXDEP];
  byte dir[BT_MAXDEP];
  size_t depth = 0;
  void *data;
  va_list argp;

  va_start(argp, key);
  data = va_arg(argp, void *);
  va_end(argp);

  tree = &btree->tree;
  path[depth] = (struct GDMbtree_node *)tree;
  dir[depth++] = PATH_LEFT;

  while(*tree != NULL) {
    int v = btree->pcmp(key, (**tree).key);
    if(v == 0) {
      if(btree->opts & GDMBT_HASH) {
        if(btree->opts & GDMBT_CPTR) {
          char *px = (**tree).data;
          HX_strclone(&px, data);
        } else {
          (**tree).data = data;
        }
      } else if(btree->opts & GDMBT_CPTR) {
        char *px = (**tree).data;
        HX_strclone(&px, key);
      }
      return *tree;
    }

    path[depth] = *tree;
    v = v > 0;
    dir[depth++] = v;
    tree = &(**tree).st[v];
  }

  if((node = *tree = malloc(sizeof(struct GDMbtree_node))) == NULL) {
    return NULL;
  }

  ++btree->itemcount;
  path[depth - 1]->st[dir[depth - 1]] = node;
  node->st[PATH_LEFT] = node->st[PATH_RIGHT] = NULL;
  node->color = NODE_RED;

  if(btree->opts & GDMBT_HASH) {
    if(btree->opts & GDMBT_LKEY) { node->key = key; }
    else { node->key = HX_strdup(key); }
    if(btree->opts & GDMBT_CPTR) { node->data = HX_strdup(data); }
    else { node->data = data; }
  } else {
    if(btree->opts & GDMBT_CPTR) { node->key = HX_strdup(key); }
    else { node->key = key; }
  }

  if(depth >= 3 && path[depth - 1]->color == NODE_RED) { ++btree->ta; }
  while(depth >= 3 && path[depth - 1]->color == NODE_RED) {
    if(dir[depth - 2] == PATH_LEFT) {
      struct GDMbtree_node *y = path[depth - 2]->st[PATH_RIGHT];
      if(y != NULL && y->color == NODE_RED) {
        path[depth - 1]->color = y->color = NODE_BLACK;
        path[depth - 2]->color = NODE_RED;
        depth -= 2;
      } else {
        struct GDMbtree_node *x;
        if(dir[depth - 1] == PATH_LEFT) {
          y = path[depth - 1];
        } else {
          x = path[depth - 1];
          y = x->st[PATH_RIGHT];
          x->st[PATH_RIGHT] = y->st[PATH_LEFT];
          y->st[PATH_LEFT] = x;
          path[depth - 2]->st[PATH_LEFT] = y;
        }

        x = path[depth - 2];
        x->st[PATH_LEFT] = y->st[PATH_RIGHT];
        y->st[PATH_RIGHT] = x;
        path[depth - 3]->st[dir[depth - 3]] = y;
        x->color = NODE_RED;
        y->color = NODE_BLACK;
        break;
      }
    } else {
      struct GDMbtree_node *y = path[depth - 2]->st[PATH_LEFT];
      if(y != NULL && y->color == NODE_RED) {
        path[depth - 1]->color = y->color = NODE_BLACK;
        path[depth - 2]->color = NODE_RED;
        depth -= 2;
      } else {
        struct GDMbtree_node *x;
        if(dir[depth - 1] == PATH_RIGHT) {
          y = path[depth - 1];
        } else {
          x = path[depth - 1];
          y = x->st[PATH_LEFT];
          x->st[PATH_LEFT] = y->st[PATH_RIGHT];
          y->st[PATH_RIGHT] = x;
          path[depth - 2]->st[PATH_RIGHT] = y;
        }

        x = path[depth - 2];
        x->st[PATH_RIGHT] = y->st[PATH_LEFT];
        y->st[PATH_LEFT] = x;
        path[depth - 3]->st[dir[depth - 3]] = y;
        x->color = NODE_RED;
        y->color = NODE_BLACK;
        break;
      }
    }
  }
  return node;
}

struct GDMbtree_node *GDMbtree_find(struct GDMbtree *btree, void *key) {
  struct GDMbtree_node *node;
  int v;
  node = btree->tree;
  while(node != NULL) {
    if((v = btree->pcmp(key, node->key)) == 0) { return node; }
    node = node->st[v > 0];
  }
  return NULL;
}

void *GDMbtree_get(struct GDMbtree *btree, void *q) {
  struct GDMbtree_node *n;
  if((n = GDMbtree_find(btree, q)) == NULL) { return NULL; }
  return n->data;
}

void *GDMbtree_del(struct GDMbtree *btree, void *key) {
  struct GDMbtree_node **tree, *p;
  struct GDMbtree_node *path[BT_MAXDEP];
  byte dir[BT_MAXDEP];
  size_t depth = 0;
  void *itemptr;
  int v;

  if(*(tree = &btree->tree) == NULL) { return NULL; }
  path[depth] = (struct GDMbtree_node *)tree;
  dir[depth++] = PATH_LEFT;

  while(*tree != NULL) {
    if((v = btree->pcmp(key, (**tree).key)) == 0) { break; }
    v = v > 0;
    path[depth] = *tree;
    dir[depth++] = v;
    tree = &(**tree).st[v];
  }

  if(*tree == NULL) { return NULL; }
  p = *tree;
  if(btree->opts & GDMBT_HASH) { itemptr = p->data; }
  else { itemptr = p->key; }

  ++btree->ta;
  if(p->st[PATH_RIGHT] == NULL) {
    path[depth - 1]->st[dir[depth - 1]] = p->st[PATH_LEFT];
  } else {
    struct GDMbtree_node *r = p->st[PATH_RIGHT];
    byte color;

    if(r->st[PATH_LEFT] == NULL) {
      r->st[PATH_LEFT] = p->st[PATH_LEFT];
      color = r->color;
      r->color = p->color;
      p->color = color;
      path[depth - 1]->st[dir[depth - 1]] = r;
      path[depth] = r;
      dir[depth++] = PATH_RIGHT;
    } else {
      struct GDMbtree_node *s;
      size_t spos = depth++;

      while(1) {
        path[depth] = r;
        dir[depth++] = PATH_LEFT;
        if((s = r->st[PATH_LEFT])->st[PATH_LEFT] == NULL) { break; }
        r = s;
      }

      path[spos] = s;
      dir[spos] = PATH_RIGHT;
      path[spos - 1]->st[dir[spos - 1]] = s;
      s->st[PATH_LEFT] = p->st[PATH_LEFT];
      r->st[PATH_LEFT] = s->st[PATH_RIGHT];
      s->st[PATH_RIGHT] = p->st[PATH_RIGHT];
      color = s->color;
      s->color = p->color;
      p->color = color;
    }
  }

  if(p->color == NODE_BLACK) {
    while(1) {
      struct GDMbtree_node *x = path[depth - 1]->st[dir[depth - 1]];
      if(x != NULL && x->color == NODE_RED) {
        x->color = NODE_BLACK;
        break;
      }

      if(depth < 2) { break; }
      if(dir[depth - 1] == PATH_LEFT) {
        struct GDMbtree_node *w = path[depth - 1]->st[PATH_RIGHT];
        if(w->color == NODE_RED) {
          w->color = NODE_BLACK;
          path[depth - 1]->color = NODE_RED;
          path[depth - 1]->st[PATH_RIGHT] = w->st[PATH_LEFT];
          w->st[PATH_LEFT] = path[depth - 1];
          path[depth - 2]->st[dir[depth - 2]] = w;
          path[depth] = path[depth - 1];
          dir[depth] = PATH_LEFT;
          path[depth - 1] = w;
          w = path[++depth - 1]->st[PATH_RIGHT];
        }

        if((w->st[PATH_LEFT] == NULL || w->st[PATH_LEFT]->color == NODE_BLACK) &&
         (w->st[PATH_RIGHT] == NULL || w->st[PATH_RIGHT]->color == NODE_BLACK)) { // L-C1
          w->color = NODE_RED;
        } else {
          if(w->st[PATH_RIGHT] == NULL || w->st[PATH_RIGHT]->color == NODE_BLACK) { // L-C3
            struct GDMbtree_node *y = w->st[PATH_LEFT];
            y->color = NODE_BLACK;
            w->color = NODE_RED;
            w->st[PATH_LEFT] = y->st[PATH_RIGHT];
            y->st[PATH_RIGHT] = w;
            w = path[depth - 1]->st[PATH_RIGHT] = y;
          }

          w->color = path[depth - 1]->color;
          path[depth - 1]->color = NODE_BLACK;
          w->st[PATH_RIGHT]->color = NODE_BLACK;
          path[depth - 1]->st[PATH_RIGHT] = w->st[PATH_LEFT];
          w->st[PATH_LEFT] = path[depth - 1];
          path[depth - 2]->st[dir[depth - 2]] = w;
          break;
        }
      } else {
        struct GDMbtree_node *w = path[depth - 1]->st[PATH_LEFT];
        if(w->color == NODE_RED) {
          w->color = NODE_BLACK;
          path[depth - 1]->color = NODE_RED;
          path[depth - 1]->st[PATH_LEFT] = w->st[PATH_RIGHT];
          w->st[PATH_RIGHT] = path[depth - 1];
          path[depth - 2]->st[dir[depth - 2]] = w;
          path[depth] = path[depth - 1];
          dir[depth] = PATH_RIGHT;
          path[depth - 1] = w;
          w = path[++depth - 1]->st[PATH_LEFT];
        }

        if((w->st[PATH_LEFT] == NULL || w->st[PATH_LEFT]->color == NODE_BLACK) && (w->st[PATH_RIGHT] == NULL || w->st[PATH_RIGHT]->color == NODE_BLACK)) { // R-C1
          w->color = NODE_RED;
        } else {
          if(w->st[PATH_LEFT] == NULL || w->st[PATH_LEFT]->color == NODE_BLACK) { // R-C3
            struct GDMbtree_node *y = w->st[PATH_RIGHT];
            y->color = NODE_BLACK;
            w->color = NODE_RED;
            w->st[PATH_RIGHT] = y->st[PATH_LEFT];
            y->st[PATH_LEFT] = w;
            w = path[depth - 1]->st[PATH_LEFT] = y;
          }

          w->color = path[depth - 1]->color;
          path[depth - 1]->color = NODE_BLACK;
          w->st[PATH_LEFT]->color = NODE_BLACK;
          path[depth - 1]->st[PATH_LEFT] = w->st[PATH_RIGHT];
          w->st[PATH_RIGHT] = path[depth - 1];
          path[depth - 2]->st[dir[depth - 2]] = w;
          break;
        }
      }
      --depth;
    }
  }

  if(btree->opts & GDMBT_HASH) {
    if(!(btree->opts & GDMBT_LKEY)) { free(p->key); }
    if(btree->opts & GDMBT_CPTR) { free(p->data); }
  } else if(btree->opts & GDMBT_CPTR) {
    free(p->key);
  }

  free(p);
  --btree->itemcount;
  return itemptr;
}

struct GDMbtrav *GDMbtrav_init(struct GDMbtree *btree, void *start) {
  struct GDMbtrav *travp;
  if((travp = calloc(1, sizeof(struct GDMbtrav))) == NULL) { return NULL; }
  travp->btree = btree;

  if(!(btree->opts & GDMBT_HASH && start == NULL)) {
    travp->cnode = GDMbtree_find(btree, start);
    if(travp->cnode != NULL) { GDMbtrav_cid(travp, travp->cnode); }
  }

  return travp;
}

struct GDMbtree_node *GDMbtraverse(struct GDMbtrav *travp) {
  if(travp->tree_id != travp->btree->ta || travp->cnode == NULL) {
    return GDMbtrav_rewalk(travp);
  }

  return GDMbtrav_next(travp);
}

void GDMbtrav_free(struct GDMbtrav *travp) {
  HX_strclone((char **)&travp->cid, NULL);
  free(travp);
  return;
}

void GDMbtree_free(struct GDMbtree *btree) {
  GDMbtree_free_dive(btree, btree->tree);
  free(btree);
  return;
}

static void GDMbtrav_cid(struct GDMbtrav *travp, struct GDMbtree_node *node) {
  struct GDMbtree *btree = travp->btree;
  if(btree->opts & GDMBT_HASH) { HX_strclone((char **)&travp->cid, node->key); }
  else { travp->cid = node->key; }
  return;
}

static struct GDMbtree_node *GDMbtrav_next(struct GDMbtrav *travp) {
  if(travp->cnode->st[PATH_RIGHT] != NULL) {
    struct GDMbtree_node *node;
    travp->dir[travp->depth++] = PATH_RIGHT;
    node = travp->cnode->st[PATH_RIGHT];

    while(node != NULL) {
      travp->path[travp->depth] = node;
      node = node->st[PATH_LEFT];
      travp->dir[travp->depth++] = PATH_LEFT;
    }
    travp->cnode = travp->path[--travp->depth];
  } else if(travp->dir[travp->depth - 1] == PATH_LEFT) {
    travp->cnode = travp->path[--travp->depth];
  } else if(travp->dir[travp->depth - 1] == PATH_RIGHT) {
    while(1) {
      if(travp->depth == 0) {
        return travp->cnode = NULL;
      }
      if(travp->dir[travp->depth - 1] != PATH_RIGHT) { break; }
      --travp->depth;
    }
    travp->cnode = travp->path[--travp->depth];
  }

  GDMbtrav_cid(travp, travp->cnode);
  return travp->cnode;
}

static struct GDMbtree_node *GDMbtrav_rewalk(struct GDMbtrav *travp) {
  struct GDMbtree *btree = travp->btree;
  struct GDMbtree_node *node = btree->tree;
  byte go_next = 0;
  travp->depth = 0;

  if(travp->cnode == NULL) {
    while(node != NULL) {
      travp->path[travp->depth] = node;
      node = node->st[PATH_LEFT];
      travp->dir[travp->depth++] = PATH_LEFT;
    }
  } else {
    struct GDMbtree_node *good = NULL, *gpath[BT_MAXDEP];
    byte gdir[BT_MAXDEP];
    byte found = 0, gdepth = 0;
    int v;

    while(node != NULL) {
      gpath[gdepth] = travp->path[travp->depth] = node;
      v = btree->pcmp(travp->cid, node->key);
      if(v == 0) {
        ++travp->depth;
        ++found;
        break;
      }
      v = v > 0;
      if(v == PATH_LEFT) {
        good = node;
        gdir[gdepth++] = v;
      }
      travp->dir[travp->depth++] = v;
      node = node->st[v];
    }

    if(found) {
      ++go_next;
    } else {
      memcpy(travp->path, gpath, sizeof(struct GDMbtree_node *) * BT_MAXDEP);
      memcpy(travp->dir, gdir, BT_MAXDEP);
      travp->depth = gdepth;
    }
  }

  travp->cnode = travp->path[--travp->depth];
  if(travp->cnode != NULL) { GDMbtrav_cid(travp, travp->cnode); }
  travp->tree_id = btree->ta;
  if(go_next) { GDMbtrav_next(travp); }
  return travp->cnode;
}

static void GDMbtree_free_dive(struct GDMbtree *btree,
 struct GDMbtree_node *node) {
  if(node == NULL) { return; }
  GDMbtree_free_dive(btree, node->st[PATH_LEFT]);
  GDMbtree_free_dive(btree, node->st[PATH_RIGHT]);
  if(btree->opts & GDMBT_HASH) {
    if(!(btree->opts & GDMBT_LKEY)) { free(node->key); }
    if(btree->opts & GDMBT_CPTR) { free(node->data); }
  } else if(btree->opts & GDMBT_CPTR) {
    free(node->key);
  }
  free(node);
  return;
}

static int GDMbtree_value_cmp(void *pa, void *pb) {
  return (pa > pb) - (pa < pb);
}

//==[ End of file ]============================================================
