/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
supply/accdbinfo.c - Info about ACCDB back-end modules
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"
#define MAXFNLEN 1024 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

static void backend_info(char *, size_t *);
static void read_ldso_conf(void);
static void read_environment(void);

static int get_options(int *, const char ***);

static struct HXlld *dirs = NULL;
int OP_open = 0;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct HXlld_node *cd;
    size_t mcount = 0;

    dirs = HXlld_init(0);
    HXlld_push(dirs, "/lib");
    HXlld_push(dirs, "/usr/lib");
    read_ldso_conf();
    HXlld_push(dirs, ".");
    read_environment();

    get_options(&argc, (const char ***)&argv);

    // "*" is not quite a module so it is NOT counted
    backend_info("*", NULL);

    cd = dirs->first;
    while(cd != NULL) {
        struct HXdir *cdp = HXdir_open(cd->ptr);
        char *dentry;
        if(cdp == NULL) { continue; }

        while((dentry = HXdir_read(cdp)) != NULL) {
            char buf[MAXFNLEN];
            if(strncmp(dentry, "accdb_", 6) != 0) { continue; }
            snprintf(buf, MAXLNLEN, "%s/%s", (char *)cd->ptr, dentry);
            backend_info(buf, &mcount);
        }

        HXdir_close(cdp);
        cd = cd->Next;
    }

    printf("Read %u modules\n", mcount);
    return !1;
}

static void backend_info(char *fn, size_t *mcount) {
    struct accdb_module *md = accdb_load(fn);
    if(md == NULL) { return; }
    char sepl = '<', sepr = '>';
    if(strcmp(fn, "*") == 0) { sepl = sepr = '"'; }

    printf("Module %c%s%c:\n", sepl, fn, sepr);
    printf("    Name       : %s\n", md->name);
    printf("    Description: %s\n", md->desc);
    if(strchr(md->info, '\n') == NULL) {
        printf("    Info       : %s\n", md->info);
    } else {
        printf("    Info:\n%s\n", md->info);
    }

    if(OP_open) {
        md->open(&md->state);
        md->close(md->state);
    }

    accdb_unload(md);
    if(mcount != NULL) { ++*mcount; }
    return;
}

static void read_ldso_conf(void) {
    FILE *fp;
    char buf[MAXLNLEN];

    if((fp = fopen("/etc/ld.so.conf", "r")) == NULL) { return; }
    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *q;
        HX_chomp(buf);
        if((q = strchr(buf, '=')) != NULL) { *q = '\0'; }
        HXlld_push(dirs, buf);
    }

    fclose(fp);
    return;
}

static void read_environment(void) {
    char *entry, *our, *travp;
    if((entry = getenv("LD_LIBRARY_PATH")) == NULL) { return; }
    HX_strclone(&our, entry);
    travp = our;

    while((travp = HX_strsep(&travp, ":")) != NULL) {
        HXlld_push(dirs, travp);
    }

    free(our);
    return;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    char *flag_l = NULL;

    const struct poptOption options_table[] = {
        {NULL, 'L', POPT_ARG_STRING, &flag_l, 'L',
         "Additional search directory", "dir"},
        {NULL, 'O', POPT_ARG_NONE, &OP_open, '\0',
         "Open (q_open) the ACCDB back-end", NULL},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            case 'L':
                HXlld_push(dirs, flag_l);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = Sh_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    poptFreeContext(ctx);
    return 1;
}

//==[ End of file ]============================================================
