/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
sysprog/userdel.c - Delete a user account and related files
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <accdb.h>
#include "sysprog/shared.h"
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
    E_SUCCESS = 0,
    E_OTHER,       // other error, see errno
    E_OPEN,        // unable to open back-end module or DB
    E_NOEXIST,     // user does not exist
    E_UPDATE,      // db->userdel() did not return ok
    E_CLOSE,       // db->close() did not return ok
};

struct {
    char *lname;
    // userdel internal
    char *ac_after, *ac_before;
    int rm_home;
} opt = {
    lname: NULL,
    ac_after: NULL, ac_before: NULL, rm_home: 0,
};

static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_user mmask;
    int eax, rv = E_SUCCESS;

    memset(&mmask, 0, sizeof(struct accdb_user));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

    // ----------------------------------------
    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if((eax = db->open(&db->state)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    memset(&mmask, 0, sizeof(struct accdb_user));
    mmask.gid = -1;
    mmask.lname = opt.lname = *argv;

    // ----------------------------------------
    if((eax = db->userinfo(db->state, &mmask, NULL, 0)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    }

    if(eax == 0) {
        fprintf(stderr, "User \"%s\" does not exist\n", mmask.lname);
        rv = E_NOEXIST;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(opt.ac_before != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_before, mmask.lname);
        system(cmd);
    }

    if((eax = db->userdel(db->state, &mmask)) <= 0) {
        fprintf(stderr, "Error: Deleting user failed: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(opt.ac_after != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_after, mmask.lname);
        system(cmd);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db->state)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed:"
         " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}


//-----------------------------------------------------------------------------
static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, accdb-userdel options
        {NULL, 'R', POPT_ARG_VAL, &opt.rm_home, 2,
         "Remove: home directory, mail spool and cron spool tabs", NULL},

        // Default options
        {NULL, 'r', POPT_ARG_NONE, &opt.rm_home, 1,
         "Remove: home directory and mail spool", NULL},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0);
    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = Sh_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(args == NULL || *args == NULL) {
        // loginname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static void read_config(void) {
    const char *paths[] = {"etc", "/etc", NULL};
    struct rconfig_opt options_table[] = {
        {"AC_BEFORE", 's', &opt.ac_before, NULL},
        {"AC_AFTER", 's', &opt.ac_after, NULL},
        {NULL},
    };

    accdb_rconfig_pv(paths, "userdel.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
