/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
sysprog/usermod.c - Modify a user account
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <accdb.h>
#include "sysprog/shared.h"
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
    E_SUCCESS = 0,
    E_OTHER,       // other error, see errno
    E_OPEN,        // unable to open back-end module or DB
    E_NOEXIST,     // user does not exist
    E_UID_USED,    // UID already used and -o was not specified
    E_NAME_USED,   // new user name (-n) already exists
    E_UPDATE,      // db->groupmod() did not return ok
    E_CLOSE,       // db->close() did not return ok
};

struct {
    char *lname, *new_lname;
    long new_uid, new_gid;
    char *new_gecos, *new_homedir, *new_shell;

    // usermod internal
    char *ac_after, *ac_before;
    int dup, inter;
} opt = {
    lname: NULL, new_lname: NULL, new_uid: -1, new_gid: -1, new_gecos: NULL,
    new_homedir: NULL, new_shell: NULL,
    ac_after: NULL, ac_before: NULL, dup: 0, inter: 0,
};

static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_user mmask, smask, result;
    int eax, rv = E_SUCCESS;

    memset(&mmask, 0, sizeof(struct accdb_user));
    memset(&smask, 0, sizeof(struct accdb_user));
    memset(&result, 0, sizeof(struct accdb_user));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return !eax; }

    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if((eax = db->open(&db->state)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    smask.lname = opt.lname = *argv;
    smask.uid = -1;
    smask.gid = -1;

    if((eax = db->userinfo(db->state, &smask, &result, 0)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    }

    if(eax == 0) {
        fprintf(stderr, "User \"%s\" does not exist\n", smask.lname);
        rv = E_NOEXIST;
      goto __main__close_backend;
    }

  // ----------------------------------------
    if(opt.inter) { memcpy(&mmask, &result, sizeof(struct accdb_user)); }
    if(opt.new_lname != NULL) { mmask.lname = opt.new_lname; }
    if(opt.new_uid != -1) { mmask.uid = opt.new_uid; }

  // ----------------------------------------
    if(opt.ac_before != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_before, mmask.lname, mmask.uid);
        system(cmd);
    }

    if((eax = db->usermod(db->state, &smask, &mmask)) <= 0) {
        fprintf(stderr, "Error: User updating failed: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(opt.ac_after != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, opt.ac_after, mmask.lname, mmask.uid);
        system(cmd);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db->state)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed:"
         " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, uudb-groupmod options
        {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_after,
         '\0', "Program to run after group modification", NULL},
        {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_before,
         '\0', "Program to run before group modification", NULL},
        {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
         "Interactively prompt for parameters", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},

        // Default options
        {NULL, 'c', POPT_ARG_STRING, &opt.new_gecos, '\0',
         "New comment for the user", "comment"},
        {NULL, 'd', POPT_ARG_STRING, &opt.new_homedir, '\0',
         "New home directory (home automatically moved)", "dir"},
        {NULL, 'g', POPT_ARG_LONG, &opt.new_gid, '\0',
         "The new primary group for the user", "group"},
        {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
         "Allow creating a user with non-unique UID"
         " (might be disabled by back-end module)", NULL},
        {NULL, 'p', POPT_ARG_STRING, NULL, '\0',
         "Password", NULL}, // UNIMPLEMENTED
        {NULL, 's', POPT_ARG_STRING, &opt.new_shell, '\0',
         "The user's shell", "prog"},
        {NULL, 'u', POPT_ARG_LONG, &opt.new_uid, '\0',
         "Numerical value of the user's ID", "uid"},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0);
    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = Sh_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(args == NULL || *args == NULL) {
        // loginname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static void read_config(void) {
    const char *paths[] = {"etc", "/etc", NULL};
    struct rconfig_opt options_table[] = {
        {"AC_BEFORE", 's', &opt.ac_before, NULL},
        {"AC_AFTER", 's', &opt.ac_after, NULL},
        {NULL},
    };

    accdb_rconfig_pv(paths, "usermod.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
